package array;
/**
 * 题目：41. 缺失的第一个正数 (firstMissingPositive)
 * 描述：给你一个未排序的整数数组 nums ，请你找出其中没有出现的最小的正整数。
 *
 * 请你实现时间复杂度为 O(n) 并且只使用常数级别额外空间的解决方案。

 *
 示例 1:
 输入：nums = [1,2,0]
 输出：3
 解释：范围 [1,2] 中的数字都在数组中。

 * 链接：https://leetcode.cn/problems/first-missing-positive/description/
 */
//不会
public class FirstMissingPositive {
    /**
     * 定位阶段
     * 遍历数组，对每个元素 nums[i]，如果它满足 1 ≤ nums[i] ≤ n，且没有放在正确的位置上（即 nums[nums[i]−1] ≠ nums[i]），就把它和目标位置上的元素交换。
     *
     * 这样做相当于把所有「有可能成为第一个缺失正数」的数放到它们理想的位置上。
     * 每个元素最多交换一次或两次，因此这一阶段仍是 O(n)。
     * 检查阶段
     * 再次遍历数组，找到第一个下标 i，使得 nums[i] ≠ i+1，此时 i+1 就是缺失的最小正数；
     * 如果遍历结束都没发现不匹配的，则说明 1…n 都出现过，答案是 n+1。
     *
     * 为什么是常数空间？
     * 我们只是在原数组上做交换，没有使用额外的数组或哈希表。
     * @param nums
     * @return
     */
    public int firstMissingPositive(int[] nums) {
        int n = nums.length;
        // 定位阶段：把每个 1...n 的数放到下标 num-1
        for (int i = 0; i < n; i++) {
            // 当 nums[i] 落在 [1, n] 且没有放在正确位置时，交换到正确位置
            while (nums[i] >= 1 && nums[i] <= n && nums[nums[i] - 1] != nums[i]) {
                swap(nums, i, nums[i] - 1);
            }
        }
        // 检查阶段：第一个 nums[i] != i+1 的下标即答案
        for (int i = 0; i < n; i++) {
            if (nums[i] != i + 1) {
                return i + 1;
            }
        }
        // 全都对上了，则缺失的是 n+1
        return n + 1;
    }

    // 交换数组中两个元素
    private void swap(int[] arr, int i, int j) {
        int tmp = arr[i];
        arr[i] = arr[j];
        arr[j] = tmp;
    }
}
