package matrix;

import java.util.HashSet;

/**
 * 题目：73. 矩阵置零 (setZeroes)
 * 描述：给定一个 m x n 的矩阵，如果一个元素为 0 ，则将其所在行和列的所有元素都设为 0 。请使用 原地 算法。

 * 链接：https://leetcode.cn/problems/set-matrix-zeroes/
 示例 1:
 输入：matrix = [[1,1,1],[1,0,1],[1,1,1]]
 输出：[[1,0,1],[0,0,0],[1,0,1]]

 示例 2:
 输入：matrix = [[0,1,2,0],[3,4,5,2],[1,3,1,5]]
 输出：[[0,0,0,0],[0,4,5,0],[0,3,1,0]]
 */
//二刷不会原地算法
public class SetZeroes {
    //非原地算法
    public void setZeroes1(int[][] matrix) {
        int row=matrix.length;
        int col=matrix[0].length;
        HashSet<Integer>iset=new HashSet<>();
        HashSet<Integer>jset=new HashSet<>();
        for (int i = 0; i < row; i++) {
            for (int j = 0; j < col; j++) {
                if(matrix[i][j]==0) {
                    iset.add(i);
                    jset.add(j);
                }
            }
        }
        for (Integer integer : iset) {
            for (int j = 0; j < col; j++) {
                matrix[integer][j]=0;
            }
        }
        for (Integer integer : jset) {
            for (int i = 0; i < row; i++) {
                matrix[i][integer]=0;
            }
        }
    }

    /**
     * 预处理第一行和第一列
     * 遍历第一行，看有没有 0，记到布尔变量 firstRowZero
     * 遍历第一列，看有没有 0，记到布尔变量 firstColZero
     * 这样做是因为我们后面要把其余位置的「标记」都写到第一行/第一列上，不能让原有的第 1 行/1 列信息丢失。
     *
     * 用第一行/列做标记
     * 从 (i=1,j=1) 开始遍历矩阵，
     * 如果 matrix[i][j] == 0，则将 matrix[i][0] = 0（标记该行要置零）和 matrix[0][j] = 0（标记该列要置零）。
     *
     * 根据标记清零
     * 再次遍历 (i=1,j=1) 区域：
     * 如果 matrix[i][0] == 0 或 matrix[0][j] == 0，就把 matrix[i][j] 设为 0。
     *
     * 处理第 1 行和第 1 列
     * 如果 firstRowZero 为 true，就整行清 0；
     * 如果 firstColZero 为 true，就整列清 0。
     * @param matrix
     */
    public void setZeroes(int[][] matrix) {
        int m = matrix.length;
        int n = matrix[0].length;
        boolean firstRowZero = false;
        boolean firstColZero = false;

        // 1. 预处理第一行
        for (int j = 0; j < n; j++) {
            if (matrix[0][j] == 0) {
                firstRowZero = true;
                break;
            }
        }

        // 1. 预处理第一列
        for (int i = 0; i < m; i++) {
            if (matrix[i][0] == 0) {
                firstColZero = true;
                break;
            }
        }

        // 2. 用第一行/列做标记
        for (int i = 1; i < m; i++) {
            for (int j = 1; j < n; j++) {
                if (matrix[i][j] == 0) {
                    matrix[i][0] = 0;
                    matrix[0][j] = 0;
                }
            }
        }

        // 3. 根据标记清零内部区域
        for (int i = 1; i < m; i++) {
            for (int j = 1; j < n; j++) {
                if (matrix[i][0] == 0 || matrix[0][j] == 0) {
                    matrix[i][j] = 0;
                }
            }
        }

        // 4. 清零第一行（如果需要）
        if (firstRowZero) {
            for (int j = 0; j < n; j++) {
                matrix[0][j] = 0;
            }
        }
        // 4. 清零第一列（如果需要）
        if (firstColZero) {
            for (int i = 0; i < m; i++) {
                matrix[i][0] = 0;
            }
        }
    }
}
