package substring;

import java.util.*;

/**
 * 题目：239. 滑动窗口最大值 (maxSlidingWindow)
 * 描述：给你一个整数数组 nums，有一个大小为 k 的滑动窗口从数组的最左侧移动到数组的最右侧。你只可以看到在滑动窗口内的 k 个数字。滑动窗口每次只向右移动一位。
 * 返回 滑动窗口中的最大值 。
 *
 * 链接：https://leetcode.cn/problems/sliding-window-maximum/
 *
 示例 1：

 输入：nums = [1,3,-1,-3,5,3,6,7], k = 3
 输出：[3,3,5,5,6,7]
 解释：
 滑动窗口的位置                最大值
 ---------------               -----
 [1  3  -1] -3  5  3  6  7       3
 1 [3  -1  -3] 5  3  6  7       3
 1  3 [-1  -3  5] 3  6  7       5
 1  3  -1 [-3  5  3] 6  7       5
 1  3  -1  -3 [5  3  6] 7       6
 1  3  -1  -3  5 [3  6  7]      7

 示例 2：
 输入：nums = [1], k = 1
 输出：[1]
 */
//普通方法超时 需重做
//二刷也没想到低复杂度方法
public class MaxSlidingWindow {
    /**
     * 思路说明：用优先队列（大顶堆）维护滑动窗口中的最大值
     * 滑动窗口长度为 k，我们每次都想拿到窗口内的最大值。
     * 用 PriorityQueue<int[]> 维护一个大顶堆：
     * 存入的是一个数组 new int[]{值, 下标}；
     * 排序规则：按值降序，值相同时按下标降序（保证堆顶是最大值的最新位置）。
     * 每加入一个新元素，我们就把它压入堆；
     * 然后检查堆顶：如果它的下标已经滑出窗口左边（≤ i - k），就移除它；
     * 每次窗口滑完一个位置后，堆顶就是当前窗口最大值。
     * @param nums
     * @param k
     * @return
     */
    public int[] maxSlidingWindow1(int[] nums, int k) {
        int n = nums.length;

        // 优先队列（大顶堆），按值降序；如果值相同，则按下标降序（保证更靠右的元素在上面）
        PriorityQueue<int[]> pq = new PriorityQueue<>(
                (a, b) -> a[0] != b[0] ? b[0] - a[0] : b[1] - a[1]
        );

        // 预处理第一个窗口：放入前 k 个元素
        for (int i = 0; i < k; ++i) {
            pq.offer(new int[]{nums[i], i});
        }

        // 初始化结果数组（共 n - k + 1 个窗口）
        int[] ans = new int[n - k + 1];
        ans[0] = pq.peek()[0]; // 第一个窗口的最大值

        // 开始滑动窗口
        for (int i = k; i < n; ++i) {
            // 把新元素加入窗口
            pq.offer(new int[]{nums[i], i});

            // 如果堆顶元素的下标不在当前窗口内，说明它已经滑出去了，弹出
            while (pq.peek()[1] <= i - k) {
                pq.poll();
            }

            // 当前窗口的最大值就是堆顶元素的值
            ans[i - k + 1] = pq.peek()[0];
        }
        return ans;
    }

    //单调队列(推荐）
    /**
     * 维护一个单调队列。
     在这个滑动窗口最大值的算法中，双端队列的两端各自承担不同的职责：
     队头（Deque Front）：
     队头始终保存当前窗口内最大元素的索引。由于我们维护队列中元素对应的数值是单调递减的，所以队头的元素必定是最大的。当窗口滑动时，如果队头的索引超出窗口范围，会将其移除，以保证队头始终代表当前窗口的最大值。

     队尾（Deque Rear）：
     队尾用于存放候选的元素索引，它们按照数值从大到小的顺序排列。在加入新元素前，会从队尾移除所有比新元素小的元素索引，因为这些较小的元素在未来的窗口中不可能成为最大值。这样既保证了队列的单调递减性质，也为未来的窗口维护了正确的候选顺序。
     */
    public int[] maxSlidingWindow(int[] nums, int k) {
        int n = nums.length;
        Deque<Integer> deque = new LinkedList<>(); // 存放下标，维护一个单调递减队列（从队头到队尾）

        // 1. 初始化前 k 个元素的单调队列
        for (int i = 0; i < k; ++i) {
            // 从队尾开始移除比当前元素小的元素下标，它们不可能成为后续最大值
            while (!deque.isEmpty() && nums[i] >= nums[deque.peekLast()]) {
                deque.pollLast();
            }
            // 当前元素下标加入队尾
            deque.offerLast(i);
        }

        // 2. 初始化结果数组（总共 n - k + 1 个窗口）
        int[] ans = new int[n - k + 1];
        // 当前窗口的最大值就是队头所指的下标对应的元素
        ans[0] = nums[deque.peekFirst()];

        // 3. 开始滑动窗口：从第 k 个元素开始往右滑
        for (int i = k; i < n; ++i) {

            // 维护队列的单调性：移除所有比当前元素小的队尾元素
            while (!deque.isEmpty() && nums[i] >= nums[deque.peekLast()]) {
                deque.pollLast();
            }

            // 把当前元素下标加入队尾
            deque.offerLast(i);

            // 如果队头的下标已经滑出窗口左边，就移除它
            while (deque.peekFirst() <= i - k) {
                deque.pollFirst();
            }

            // 队头就是当前窗口的最大值
            ans[i - k + 1] = nums[deque.peekFirst()];
        }

        return ans;
    }

}
